"""Library functions for YamlTable."""


import pathlib
from typing import Any, Iterable, List, Optional, Tuple

import fastjsonschema
import yaml
from yamltable.typing import Row, Schema


def read(file_path: pathlib.Path) -> Tuple[List[Row], Optional[Schema]]:
    """Read data from YAML file.

    Args:
        file_path: YAML file path

    Raises:
        FileNotFoundError: if unable to find file path
        TypeError: if file is not organized as a list

    Return:
        YAML data
    """

    try:
        with open(file_path, "r") as handle:
            data = yaml.safe_load(handle)
    except (yaml.scanner.ScannerError, yaml.parser.ParserError) as xcpt:
        raise TypeError(f"invalid YAML file: {xcpt}")

    if isinstance(data, list):
        return data, None
    elif isinstance(data, dict):
        try:
            return data["rows"], data["schema"]
        except KeyError:
            raise TypeError("YAML file does not have a schema and rows organization")
    else:
        raise TypeError("YAML file is not organized in a tabular format")


def search(key: str, val: Any, rows: Iterable[Row]) -> List[Row]:
    """Search dictionaries by key and value.

    Args:
        key: search key
        val: key comparison value
        rows: dictionaries to search

    Return:
        matching dictionaries
    """

    return [row for row in rows if key in row and row[key] == val]


def sort(key: str, rows: Iterable[Row]) -> List[Row]:
    """Sort dictionaries based on value for supplied key name.

    Args:
        key: search key
        rows: dictionaries to sort

    Return:
        list of sorted dictionaries
    """

    return sorted(rows, key=lambda row: row[key])


def validate(rows: Iterable[Row], schema: Optional[Schema]) -> Tuple[bool, int, str]:
    """Check that each row satisfies the schema.

    Args:
        rows: dictionaries to validate
        schema: JSON schema for validation

    Return:
        whether all rows are valid, invalid row index or -1, invalid error message
    """

    try:
        validator = fastjsonschema.compile(schema)
    except (fastjsonschema.JsonSchemaDefinitionException, TypeError) as xcpt:
        return False, -1, f"invalid schema: {xcpt}"

    for idx, row in enumerate(rows):
        try:
            validator(row)
        except fastjsonschema.JsonSchemaException as xcpt:
            return False, idx, xcpt.message

    return True, -1, ""


def write(
    file_path: pathlib.Path,
    rows: List[Row],
    schema: Optional[Schema] = None,
    sort_keys: bool = False,
) -> None:
    """Write data to YAML file.

    Args:
        file_path: YAML file path
        rows: list of dictionaries to write
        schema: JSON schema dictionary
        sort_keys: whether to sort row keys
    """

    if schema is None:
        with open(file_path, "w") as handle:
            yaml.dump(rows, handle, sort_keys=sort_keys)
    else:
        data = {"schema": schema, "rows": rows}
        with open(file_path, "w") as handle:
            yaml.dump(data, handle, sort_keys=sort_keys)



"""Command line interface for YamlTable."""


import enum
import pathlib
import pdb
import pprint
from typing import List, Optional, Tuple

import typer

import yamltable
from yamltable.typing import Row, Schema


FileArg = typer.Argument(..., dir_okay=False, exists=True, file_okay=True, resolve_path=True)


app = typer.Typer()


class Code(enum.Enum):
    """Exit code statuses."""

    SUCCESS = 0
    INVALID = 1
    ERROR = 2


class Debugger:
    """Context for running interactive debug sessions."""

    state = False

    @classmethod
    def launch(cls) -> None:
        """Launch interactive debug session if state is True."""

        if cls.state:
            typer.secho("Launching interactive debug session...")
            pdb.set_trace()


class Msg(enum.Enum):
    """Colors for message types."""

    EMPTY = typer.colors.YELLOW
    ERROR = typer.colors.RED
    SUCCESS = typer.colors.BRIGHT_GREEN


@app.command()
def list(key: str, file_path: pathlib.Path = FileArg) -> None:
    """List all dictionary KEY values in FILE_PATH."""

    Debugger.launch()

    rows, _ = load_data(file_path)
    for idx, row in enumerate(rows):
        try:
            typer.echo(row[key])
        except KeyError:
            typer.secho(f"error: row {idx} does not have key {key}", fg=Msg.ERROR.value)
            typer.Exit(code=Code.ERROR.value)
            break


def load_data(file_path: pathlib.Path) -> Tuple[List[Row], Optional[Schema]]:
    """Attempt to load data from YAML file.

    Args:
        file_path: YAML file path

    Return:
        YAML row data, YAML schema
    """

    try:
        return yamltable.read(file_path)
    except (FileNotFoundError, TypeError) as xcpt:
        typer.secho(xcpt, fg=Msg.ERROR.value)
        raise typer.Exit(code=Code.ERROR.value)


@app.callback()
def main(debug: bool = typer.Option(False, help="Run with interactive debug session.")) -> None:
    """Utility for working with YAML files organized similar to a relational database table."""

    Debugger.state = debug


@app.command()
def search(key: str, value: str, file_path: pathlib.Path = FileArg) -> None:
    """Search for all dictionaries in FILE_PATH with matching KEY and VALUE pairs."""

    Debugger.launch()

    rows, _ = load_data(file_path)

    count = 0
    for match in yamltable.search(key, value, rows):
        count += 1
        typer.echo(pprint.pformat(match, indent=2))

    if count == 0:
        typer.secho(
            f"no dictionaries found with (key={key}, value={value}) pair", fg=Msg.EMPTY.value
        )


@app.command()
def sort(key: str, file_path: pathlib.Path = FileArg) -> None:
    """Sort dictionaries in FILE_PATH by KEY values."""

    Debugger.launch()

    rows, schema = load_data(file_path)
    try:
        sorted_rows = yamltable.sort(key, rows)
    except TypeError as xcpt:
        typer.secho(f"error: {xcpt}", fg=Msg.ERROR.value)
        typer.Exit(code=Code.ERROR.value)
    else:
        yamltable.write(file_path, sorted_rows, schema)


@app.command()
def validate(file_path: pathlib.Path = FileArg) -> None:
    """Check that every dictionary in FILE_PATH has conforms to its schema."""

    Debugger.launch()

    rows, schema = load_data(file_path)
    valid, row, msg = yamltable.validate(rows, schema)
    if valid:
        typer.secho("YAML file rows conform to its schema", fg=Msg.SUCCESS.value)
    elif row == -1:
        typer.secho(f"invalid schema: {msg}", fg=Msg.ERROR.value)
        typer.Exit(code=Code.ERROR.value)
    else:
        typer.secho(f"invalid row {row}: {msg}", fg=Msg.ERROR.value)
        typer.Exit(code=Code.INVALID.value)


if __name__ == "__main__":
    app()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['yamltable']

package_data = \
{'': ['*']}

install_requires = \
['fastjsonschema>=2.14.2,<3.0.0', 'pyyaml>=5.2,<6.0', 'typer>=0.0.8,<0.0.9']

entry_points = \
{'console_scripts': ['yamltable = yamltable.main:app']}

setup_kwargs = {
    'name': 'yamltable',
    'version': '0.0.5',
    'description': 'Command line utility for list organized YAML files.',
    'long_description': '# YamlTable \n\n![](https://img.shields.io/pypi/v/yamltable)\n![](https://img.shields.io/github/license/wolfgangwazzlestrauss/yamltable)\n![](https://github.com/wolfgangwazzlestrauss/yamltable/workflows/build/badge.svg)\n![](https://img.shields.io/github/repo-size/wolfgangwazzlestrauss/yamltable)\n\nYamlTable is a Python command line utility for working with YAML files organized similar to a\nrelational database table. It supports YAML files organized as a list of dictionaries, which share\nkey names and value types. YamlTable provides commands for listing, searching, sorting, etc. data\nfrom the supported files.\n\n\n## Supported YAML File Organizations\n\nYamlTable works with YAML files organized as a list of dictionaries with similar key names and value types.\n```yaml\n- name: awscli\n  description: Amazon Web Services command line client\n  website: https://aws.amazon.com/\n- name: glances\n  description: operating system monitoring interface\n  website: https://github.com/nicolargo/glances\n```\n\nThe JSON schema support is included for YAML files organized as:\n```yaml\nschema:\n  $schema: http://json-schema.org/draft-07/schema#\n  description: pipx package metadata schema\n  type: object\n  properties:\n    name:\n      type: string\n      pattern: "^[\\\\w-]+$"\n    description:\n      type: string\n    website:\n      type: string\n  required:\n  - name\n  - description\n  - website\n  additionalProperties: false\nrows:\n- name: awscli\n  description: Amazon Web Services command line client\n  website: https://aws.amazon.com/\n- name: glances\n  description: operating system monitoring interface\n  website: https://github.com/nicolargo/glances\n```\n\n\n## Getting Started\n\n### Installation\n\nYamlTable can be installed for Python 3.7+ with `pipx`.\n```bash\npipx install yamltable\n```\n\nTo reuse its library functions install with `pip`.\n```bash\npip install --user yamltable\n```\n\n\n### Commands\n\nYamlTable provides the following commands for working with YAML files:\n* `list`: list dictionary key values\n* `search`: search dictionaries by key and value\n* `sort`: sort dictionaries by key and value\n* `validate`: validate that dictionaries conform to the given JSON schema\n\n\n## Contributing\n\nSince YamlTable is in an early development phase, it is not currently open to contributors.\n\n\n## License\n\nLicensed under the [MIT](license.txt) license.\n\n',
    'author': 'Macklan Weinstein',
    'author_email': 'wolfgangwazzlestrauss@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wolfgangwazzlestrauss/yamltable',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Command line interface for sorting YAML files."""

from yamltable.lib import read, search, sort, validate, write

__all__ = ["read", "search", "sort", "validate", "write"]
__author__ = "Macklan Weinstein"
__version__ = "0.0.5"
